﻿-- Название мода: TerrainHeightAdjuster
-- Версия: 1.0.0.5
-- Автор: Erata

-- Регистрация мода как специализации
TerrainHeightAdjuster = {}
local TerrainHeightAdjuster_mt = Class(TerrainHeightAdjuster, Mission00)

-- Включение режима отладки
local debugMode = true -- Установите true для вывода сообщений в консоль

-- Тестовое замедление ускорения
local testSlowestAcceleration = false -- Установите false, чтобы отключить
local testSlowdownFactor = 5.0 -- Множитель для экстремального замедления

-- Глобальные настройки поверхности
local SurfaceSettings = {
    plowed = { friction = 0.2, resistance = 1.8, deformation = 1.0, stuckChance = 0.5, slipChance = 0.8, moistureLevel = 0.2, moistureMax = 0.8 },
    cultivated = { friction = 0.3, resistance = 1.5, deformation = 0.8, stuckChance = 0.3, slipChance = 0.6, moistureLevel = 0.2, moistureMax = 0.7 },
    mud = { friction = 0.1, resistance = 2.0, deformation = 1.0, stuckChance = 0.4, slipChance = 0.5, moistureLevel = 0.5, moistureMax = 1.0 },
    grass = { friction = 0.3, resistance = 1.5, deformation = 0.6, stuckChance = 0.1, slipChance = 0.2, moistureLevel = 0.1, moistureMax = 0.5 },
    asphalt = { friction = 0.8, resistance = 1.1, deformation = 0.2, stuckChance = 0.0, slipChance = 0.0, moistureLevel = 0.0, moistureMax = 0.2 },
    dirt_road = { friction = 0.7, resistance = 1.4, deformation = 0.4, stuckChance = 0.2, slipChance = 0.3, moistureLevel = 0.2, moistureMax = 0.6 },
    natural_cover = { friction = 0.6, resistance = 1.3, deformation = 0.5, stuckChance = 0.2, slipChance = 0.4, moistureLevel = 0.2, moistureMax = 0.5 },
    default = { friction = 0.8, resistance = 1.1, deformation = 0.2, stuckChance = 0.1, slipChance = 0.2, moistureLevel = 0.1, moistureMax = 0.4 }
}

-- Временные параметры тестирования
local testingMode = false
local testSlipChance = 0.9
local testStuckChance = 0.9

-- Инициализация мода
function TerrainHeightAdjuster.initSpecialization()
    if debugMode then
        print("TerrainHeightAdjuster: Initializing specialization")
    end
end

-- Регистрация мода
function TerrainHeightAdjuster.registerSpecialization()
    SpecializationUtil.registerSpecialization("terrainHeightAdjuster", "TerrainHeightAdjuster", TerrainHeightAdjuster)
end

-- Регистрация событий
function TerrainHeightAdjuster.registerEventListeners(vehicleType)
    SpecializationUtil.registerEventListener(vehicleType, "onLoad", TerrainHeightAdjuster)
    SpecializationUtil.registerEventListener(vehicleType, "onUpdate", TerrainHeightAdjuster)
    SpecializationUtil.registerEventListener(vehicleType, "onUpdateTick", TerrainHeightAdjuster)
end

-- Загрузка мода для транспортного средства
function TerrainHeightAdjuster:onLoad(savegame)
    self.terrainHeightAdjuster = {}
    if debugMode then
        print("TerrainHeightAdjuster: Loaded for vehicle " .. tostring(self:getName()))
    end
end

-- Обновление влажности поверхности
function TerrainHeightAdjuster:updateMoistureLevel(dt, isRaining)
    local timeSinceLastRain = g_currentMission.environment.weather.timeSinceLastRain or 0
    local dryingFactor = math.clamp(timeSinceLastRain / 1440, 1, 3)
    for _, settings in pairs(SurfaceSettings) do
        if isRaining then
            local rainIntensity = g_currentMission.environment.weather:getRainFallScale() or 0
            settings.moistureLevel = math.min(settings.moistureLevel + (rainIntensity * 0.0001 * dt), settings.moistureMax)
        else
            settings.moistureLevel = math.max(settings.moistureLevel - (0.00005 * dt * dryingFactor), 0.0)
        end
    end
end

-- Изменение параметров деформации почвы
function TerrainHeightAdjuster:adjustTerrainDisplacement(isRaining, vehicle)
    local weightFactor = vehicle:getMass() / 6000
    local rainIntensityFactor = isRaining and g_currentMission.environment.weather:getRainFallScale() or 0
    local surfaceType = self:getSurfaceType(vehicle) or "default"
    local settings = SurfaceSettings[surfaceType]
    local deformationFactor = settings.deformation
    local moistureFactor = settings.moistureLevel / settings.moistureMax

    local maxDisplacement = 0.7 + (weightFactor * 0.9) + (rainIntensityFactor * 0.9) + deformationFactor + (moistureFactor * 0.5)

    if vehicle.spec_wheels then
        for _, wheel in pairs(vehicle.spec_wheels.wheels) do
            if wheel.width < 0.3 then -- Узкие шины
                maxDisplacement = maxDisplacement + 0.7
            elseif wheel.width < 0.5 then -- Стандартные шины
                maxDisplacement = maxDisplacement + 0.5
            elseif wheel.width < 0.7 then -- Широкие шины
                maxDisplacement = maxDisplacement + 0.3
            else -- Двойные шины
                maxDisplacement = maxDisplacement + 0.2
            end
        end
    end

    if debugMode then
        print(string.format("Displacement for %s: %.2f (surface=%s, raining=%s, moisture=%.2f)",
            tostring(vehicle:getName()), maxDisplacement, surfaceType, tostring(isRaining), settings.moistureLevel))
    end
end

-- Изменение мягкости почвы
function TerrainHeightAdjuster:adjustFieldSoftness(isRaining, vehicle)
    local surfaceType = self:getSurfaceType(vehicle) or "default"
    local settings = SurfaceSettings[surfaceType]
    local baseSoftness = isRaining and 0.1 or 0.4
    local weightFactor = vehicle:getMass() / 6000
    local moistureFactor = settings.moistureLevel / settings.moistureMax
    local resistance = baseSoftness - (weightFactor * 0.3) - (moistureFactor * 0.2)

    if vehicle.spec_wheels then
        for _, wheel in pairs(vehicle.spec_wheels.wheels) do
            wheel.groundFrictionCoefficient = math.max(resistance, 0.05)
        end
    end

    if debugMode then
        print(string.format("Field softness for %s: %.2f (moisture=%.2f)", tostring(vehicle:getName()), resistance, settings.moistureLevel))
    end
end

-- Применение силы тяжести, инерции, регулировка разгона
function TerrainHeightAdjuster:applyGravityAndInertia(vehicle, isRaining)
    if not vehicle.spec_motorized or not vehicle.spec_motorized.motor or not vehicle.spec_drivable then
        return
    end

    local mass = vehicle:getMass() or 3000
    local gravity = 9.8
    local surfaceType = self:getSurfaceType(vehicle) or "default"
    local settings = SurfaceSettings[surfaceType]
    local moistureFactor = settings.moistureLevel / settings.moistureMax

    local baseAcceleration = vehicle.spec_motorized.motor:getAccelerationLimit() or 1.0
    local maxAcceleration = baseAcceleration

    local currentSpeed = vehicle.spec_drivable.lastSpeed * 3600 / 1000
    local maxSpeed = vehicle.spec_motorized.motor.maxForwardSpeed or 20
    local speedRatio = math.max(0, math.min(currentSpeed / maxSpeed, 1.0))

    local massFactor = math.max(1.0, mass / 3000)
    local accelerationReduction = 0.5 / massFactor

    local surfaceResistance = settings.resistance * (1.0 + moistureFactor * 0.5)
    if isRaining then
        surfaceResistance = surfaceResistance * (1.0 + 0.3 * (g_currentMission.environment.weather:getRainFallScale() or 0))
    end
    accelerationReduction = accelerationReduction * (1.0 + surfaceResistance * 0.2)

    local loadWeight = vehicle.spec_fillUnit and vehicle.spec_fillUnit.fillUnits[1].fillLevel or 0
    local loadFactor = 1.0 + (loadWeight / 5000)
    accelerationReduction = accelerationReduction * loadFactor

    local implementFactor = 1.0
    if vehicle.spec_attacherJoints and vehicle.spec_attacherJoints.attachedImplements then
        for _, implement in pairs(vehicle.spec_attacherJoints.attachedImplements) do
            if implement.object.spec_plow then
                implementFactor = implementFactor + 0.3
            elseif implement.object.spec_cultivator then
                implementFactor = implementFactor + 0.2
            elseif implement.object.spec_sowingMachine then
                implementFactor = implementFactor + 0.1
            end
        end
    end
    accelerationReduction = accelerationReduction * implementFactor

    local inertiaFactor = 0.05 * mass
    accelerationReduction = accelerationReduction + (inertiaFactor * gravity * 0.005)

    local inertiaMultiplier = 1.0 + (speedRatio * 0.5)
    if speedRatio < 0.2 then
        inertiaMultiplier = 0.8
    end
    accelerationReduction = accelerationReduction * inertiaMultiplier

    local brakingFactor = 1.0
    if vehicle.spec_drivable.isBraking then
        brakingFactor = 1.0 + (speedRatio * 0.7)
        if settings.slipChance > 0.4 or surfaceType == "mud" then
            brakingFactor = brakingFactor * 1.2
        end
        accelerationReduction = accelerationReduction * brakingFactor
    end

    local powerFactor = 1.0
    local torque = vehicle.spec_motorized.motor.maxTorque or 500
    local maxRpm = vehicle.spec_motorized.motor.maxRpm or 2200
    if torque and maxRpm then
        powerFactor = (torque * maxRpm) / (500 * 2200)
        powerFactor = math.max(0.5, math.min(powerFactor, 2.0))
    else
        local maxPower = vehicle.spec_motorized.motor.maxForwardPower or 100
        powerFactor = maxPower / 100
        powerFactor = math.max(0.5, math.min(powerFactor, 2.0))
    end

    local currentRpm = vehicle.spec_motorized.motor.currentRpm or maxRpm
    local rpmFactor = math.min(currentRpm / (0.5 * maxRpm), 1.0)
    powerFactor = powerFactor * (0.7 + 0.3 * rpmFactor)

    accelerationReduction = accelerationReduction / powerFactor

    if testSlowestAcceleration then
        accelerationReduction = accelerationReduction * testSlowdownFactor
    end

    local newAcceleration = math.max(maxAcceleration * (1.0 / accelerationReduction), 0.05)
    vehicle.spec_motorized.motor:setAccelerationLimit(newAcceleration)

    -- Дым при низкой мощности и тяжёлой навеске
    if powerFactor < 0.8 and implementFactor > 1.2 and vehicle.spec_exhaustEffects then
        for _, effect in pairs(vehicle.spec_exhaustEffects.exhaustEffects or {}) do
            if effect.particleSystem then
                effect.particleSystem:start()
            end
        end
    elseif vehicle.spec_exhaustEffects then
        for _, effect in pairs(vehicle.spec_exhaustEffects.exhaustEffects or {}) do
            if effect.particleSystem then
                effect.particleSystem:stop()
            end
        end
    end

    if debugMode then
        print(string.format("Acceleration for %s: mass=%.0f, surface=%s, load=%.0f, speedRatio=%.2f, implementFactor=%.2f, powerFactor=%.2f, torque=%.0f, maxRpm=%.0f, rpmFactor=%.2f, testSlowdown=%.2f, moisture=%.2f, newAccel=%.2f",
            tostring(vehicle:getName()), mass, surfaceType, loadWeight, speedRatio, implementFactor, powerFactor, torque, maxRpm, rpmFactor, testSlowestAcceleration and testSlowdownFactor or 1.0, settings.moistureLevel, newAcceleration))
    end
end

-- Учёт угла наклона
function TerrainHeightAdjuster:adjustMovementOnSlope(vehicle)
    if not vehicle.spec_drivable or not vehicle.spec_wheels then
        return
    end

    local x, y, z = getWorldTranslation(vehicle.rootNode)
    local dirX, dirY, dirZ = localDirectionToWorld(vehicle.rootNode, 0, 0, 1)
    local forwardX, forwardZ = x + dirX * 2, z + dirZ * 2
    local rearX, rearZ = x - dirX * 2, z - dirZ * 2

    local heightForward = getTerrainHeightAtWorldPos(g_currentMission.terrainRootNode, forwardX, 0, forwardZ)
    local heightRear = getTerrainHeightAtWorldPos(g_currentMission.terrainRootNode, rearX, 0, rearZ)

    local heightDiff = heightForward - heightRear
    local slopeAngle = math.deg(math.atan(heightDiff / 4))

    local mass = vehicle:getMass() or 5000
    local gravityForce = mass * 9.8 * math.sin(math.rad(slopeAngle))

    local currentSpeed = vehicle.spec_drivable.lastSpeed * 3600 / 1000
    local newSpeed = currentSpeed
    local maxSpeed = vehicle.spec_motorized and vehicle.spec_motorized.motor.maxForwardSpeed or 20

    local isRaining = g_currentMission.environment.weather:getIsRaining()
    local surfaceType = self:getSurfaceType(vehicle) or "default"
    local settings = SurfaceSettings[surfaceType]
    local slipFactor = settings.slipChance * (1.0 + (settings.moistureLevel / settings.moistureMax) * 0.5)

    if isRaining then
        slipFactor = slipFactor * (1.0 + 0.5 * (g_currentMission.environment.weather:getRainFallScale() or 0))
    end

    if slopeAngle > 0 then
        newSpeed = currentSpeed - (gravityForce * 0.0001)
        newSpeed = math.max(newSpeed, 0)
    elseif slopeAngle < 0 then
        newSpeed = currentSpeed + (math.abs(gravityForce) * 0.0001)
        if slipFactor > 0.4 or surfaceType == "mud" then
            local randomSlip = (math.random() - 0.5) * 0.2 * slipFactor
            newSpeed = newSpeed * (1.0 + randomSlip)
            if vehicle.spec_wheels then
                for _, wheel in pairs(vehicle.spec_wheels.wheels) do
                    local originalFriction = wheel.groundFrictionCoefficient or settings.friction
                    wheel.groundFrictionCoefficient = math.max(originalFriction * (1.0 - slipFactor * 0.5), 0.05)
                end
            end
        end
        newSpeed = math.min(newSpeed, maxSpeed * 1.5)
    end

    vehicle.spec_drivable.lastSpeed = newSpeed * 1000 / 3600

    if debugMode then
        print(string.format("Slope for %s: slopeAngle=%.2f, surface=%s, slipFactor=%.2f, moisture=%.2f, newSpeed=%.2f",
            tostring(vehicle:getName()), slopeAngle, surfaceType, slipFactor, settings.moistureLevel, newSpeed))
    end
end

-- Учёт загруженности
function TerrainHeightAdjuster:adjustHandlingBasedOnLoad(vehicle)
    local loadWeight = vehicle.spec_fillUnit and vehicle.spec_fillUnit.fillUnits[1].fillLevel or 0
    local baseTurningRadius = 5.0
    local adjustedTurningRadius = baseTurningRadius + loadWeight / 1000

    if vehicle.spec_drivable then
        vehicle.spec_drivable.maxTurningRadius = adjustedTurningRadius
    end

    if debugMode then
        print(string.format("Handling for %s: turning radius=%.2f", tostring(vehicle:getName()), adjustedTurningRadius))
    end
end

-- Изменение трения колёс
function TerrainHeightAdjuster:adjustWheelFrictionBasedOnWeatherAndSurface(vehicle, isRaining)
    local surfaceType = self:getSurfaceType(vehicle) or "default"
    local settings = SurfaceSettings[surfaceType]
    local friction = settings.friction
    local moistureFactor = settings.moistureLevel / settings.moistureMax
    local timeSinceLastRain = g_currentMission.environment.weather.timeSinceLastRain or 0
    local dryingFactor = math.clamp(timeSinceLastRain / 1440, 1, 3)

    -- Учёт ширины шин и массы
    local mass = vehicle:getMass() or 3000
    local numWheels = vehicle.spec_wheels and #vehicle.spec_wheels.wheels or 4
    local widthToMassRatio = 1.0

    if vehicle.spec_wheels then
        for _, wheel in pairs(vehicle.spec_wheels.wheels) do
            local wheelWidth = wheel.width or 0.5
            widthToMassRatio = math.min(wheelWidth / (mass / numWheels), 1)
            friction = friction / (1.5 - math.min(wheelWidth, 1))

            -- Коррекция трения по погоде и влажности
            if isRaining then
                local rainIntensity = g_currentMission.environment.weather:getRainFallScale() or 0
                if surfaceType == "dirt_road" then
                    friction = friction * 0.7 * (1.0 - moistureFactor * 0.3)
                elseif surfaceType == "natural_cover" then
                    friction = friction * 0.6 * (1.0 - moistureFactor * 0.2)
                else
                    friction = friction * (1.0 - 0.4 * rainIntensity) * (1.0 - moistureFactor * 0.3)
                end
            else
                if surfaceType == "dirt_road" then
                    friction = friction * 0.9 * (1.0 - moistureFactor * 0.1)
                elseif surfaceType == "natural_cover" then
                    friction = friction * 0.8 * (1.0 - moistureFactor * 0.1)
                else
                    friction = friction * (1.0 - moistureFactor * 0.1)
                end
                friction = friction / dryingFactor
            end

            -- Применение трения
            wheel.groundFrictionCoefficient = math.max(friction, 0.05)
        end
    end

    --[[
    -- Поддержка снега (закомментировано, включите, если нужна)
    if vehicle.spec_wheels and vehicle.hasSnowContact then
        local snowFactor = vehicle.snowHeight and (1 + vehicle.snowHeight * 0.33) or 1
        for _, wheel in pairs(vehicle.spec_wheels.wheels) do
            friction = settings.friction * snowFactor / dryingFactor
            if mass < 8000 then
                if widthToMassRatio > 0.06 and widthToMassRatio < 0.12 then
                    friction = friction * (1 + (wheel.width / 5))
                else
                    friction = friction * (1 - (wheel.width / 5))
                end
            end
            wheel.groundFrictionCoefficient = math.max(friction, 0.05)
        end
    end
    --]]

    if debugMode then
        print(string.format("Friction for %s: %.2f (surface=%s, moisture=%.2f, widthToMass=%.2f, timeSinceRain=%.0f)",
            tostring(vehicle:getName()), friction, surfaceType, settings.moistureLevel, widthToMassRatio, timeSinceLastRain))
    end
end

-- Симуляция пробуксовки
function TerrainHeightAdjuster:simulateCompleteSlipInsteadOfStuck(vehicle)
    if vehicle.spec_wheels then
        for _, wheel in pairs(vehicle.spec_wheels.wheels) do
            wheel.groundFrictionCoefficient = 0.05
        end
    end
    if vehicle.spec_drivable then
        vehicle.spec_drivable.lastSpeed = 0
    end

    if debugMode then
        print("Vehicle " .. tostring(vehicle:getName()) .. " is experiencing full slip!")
    end
end

-- Проверка пробуксовки
function TerrainHeightAdjuster:checkVehicleSlipInsteadOfStuck(vehicle, isRaining)
    local surfaceType = self:getSurfaceType(vehicle) or "default"
    local settings = SurfaceSettings[surfaceType]
    local stuckChance = testingMode and testStuckChance or settings.stuckChance
    local moistureFactor = settings.moistureLevel / settings.moistureMax

    if isRaining then
        local rainIntensity = g_currentMission.environment.weather:getRainFallScale() or 0
        stuckChance = stuckChance * (1.0 + 0.5 * rainIntensity) * (1.0 + moistureFactor * 0.5)
    else
        stuckChance = stuckChance * (1.0 + moistureFactor * 0.3)
    end

    if vehicle.spec_wheels and vehicle.spec_attacherJoints then
        for _, wheel in pairs(vehicle.spec_wheels.wheels) do
            if wheel.width < 0.3 and vehicle.spec_attacherJoints.attachedImplements then
                stuckChance = stuckChance + 0.2
            end
        end
    end

    if math.random() < stuckChance then
        self:simulateCompleteSlipInsteadOfStuck(vehicle)
    end
end

-- Получение типа поверхности
function TerrainHeightAdjuster:getSurfaceType(vehicle)
    local x, y, z = getWorldTranslation(vehicle.rootNode)
    local groundType = g_currentMission.fieldGroundSystem:getFieldGroundType(x, z)
    if groundType == FieldGroundType.PLOWED then
        return "plowed"
    elseif groundType == FieldGroundType.CULTIVATED then
        return "cultivated"
    elseif groundType == FieldGroundType.GRASS then
        return "grass"
    else
        return "default"
    end
end

-- Основной цикл обновления
function TerrainHeightAdjuster:onUpdate(dt)
    local isRaining = g_currentMission.environment.weather:getIsRaining()
    self:updateMoistureLevel(dt, isRaining)
    self:adjustWheelFrictionBasedOnWeatherAndSurface(self, isRaining)
    self:applyGravityAndInertia(self, isRaining)
    self:adjustMovementOnSlope(self)
    self:adjustHandlingBasedOnLoad(self)
    self:checkVehicleSlipInsteadOfStuck(self, isRaining)
    self:adjustTerrainDisplacement(isRaining, self)
    self:adjustFieldSoftness(isRaining, self)
end

-- Регистрация мода в игре
addModEventListener(TerrainHeightAdjuster)

-- Вывод отладочной информации
if debugMode then
    print("TerrainHeightAdjuster mod loaded successfully")
end